/*
  Xquote Copyright (c)1997-99 Mark Buser,
  All Rights Reserved.

  Permission is hereby granted to copy
  and freely distribute copies of this
  program for non-commercial purposes
  without fee, provided that this notice
  appears in all copies.
  
  All redistributions must be in their
  entirety as originally distributed.
  Feel free to modify Xquote, but
  modified versions may not be distributed
  without prior consent of the author.
  
  This software is provided 'as-is'
  without any express or implied warranty.
  In no event will the author be held
  liable for any damages resulting from
  the use of this software.

  $Revision: 1.9 $ $Date: 1999/12/20 23:27:18 $
*/

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include <Xm/XmAll.h>

#include "optdetail.h"
#include "parse.h"
#include "pixmap.h"
#include "server.h"
#include "status.h"
#include "tape.h"
#include "xquote.h"
#include "xutil.h"

/* Externs */
extern Widget Toplevel;

typedef struct tick_struct {
  char *name;
  int type;
  int server;
  QUERY_STRUCT *detail;
} TICK_STRUCT;

/* Globals */
static XmStringTable save;
static int  saveCount = 0;
static TICK_STRUCT **tickTable;
static int numTick = 0;
static int curType = 0;
static int curServer = 0;

#define TICKFIELDS 5

/* Forward declarations */
void tickMakeTypeMenu (Widget, int);


/*
** Ticker Symbol Support
*/

/* 
** Session save and restore functions 
*/
char *tickGetItem (int pos)
{
  char *item;
  char ticker[7];
  char type[16];
  char server[16];
  char *next;

  if (pos >= 0 && pos < numTick) {
    XmStringGetLtoR ( save[pos], XmFONTLIST_DEFAULT_TAG, &item);

    sscanf (item, "%6s =>%15c @%15c", ticker, type, server);
    type[15] = server[15] = '\0';     /* sscanf does not null terminate %c */
    if ((next = strrchr (type, ' '))) /* Prune trailing white space        */
      while (isspace(*next)) 
        *next-- = '\0';
    if ((next = strrchr (server, ' ')))
      while (isspace(*next)) 
        *next-- = '\0';

    XtFree (item);
    item = XtCalloc ( 80, sizeof (char) );
    sprintf ( item, "%s,%s,%s", ticker, type, server); 
  }
  return (item);   
}


void tickAddItem (char *item)
{
  char line[80];
  char *ticker;
  char *type;
  char *server;

  save = (XmStringTable) XtRealloc ( (char *)save, 
		                     (saveCount+1) * sizeof(XmString) );
  if (save) {
    /* Break into ticker, type, and server */
    ticker = strtok (item, ",");
    type = strtok (NULL, ",");
    server = strtok (NULL, ",");
    if (ticker && type && server) {
      sprintf (line, "%-6.6s =>%-15.15s @%-15.15s", ticker, type, server);
      save[saveCount++] = XmStringCreateLocalized (line);
    }
  }
}

int tickGetNum ()
{
  return (numTick);
}

/* Get saved ticker server */
int tickGetServer (int which)
{
  if (which < numTick)
    return ( tickTable[which]->server );
  else
    return (0);
}

/* Get saved ticker type */
int tickGetType ( int which )
{
  if (which < numTick)
    return ( tickTable[which]->type );
  else
    return (0);
}

/* Get saved ticker name */
char *tickGetName ( int which )
{
  if (which < numTick)
    return ( tickTable[which]->name );
  else
    return ((char *)NULL);
}

/* Get saved data retrieved from net for displaying */
QUERY_STRUCT *tickGetDetail ( int which )
{
  if (which < numTick)
    return ( tickTable[which]->detail );
  else
    return ((QUERY_STRUCT *)NULL);
}


/* Save data retrieved from net for displaying */
void tickSetDetail ( int which, QUERY_STRUCT *data )
{

  /* Update the data structure */
  if ( which < numTick  ) {                   /* Replace existing */
    freePattern (tickTable[which]->detail);
    tickTable[which]->detail = data;
  } 
  
  /* Update the main window display */
}

/* Add everthing from the tick editor list to the main window list */
void tickAddMainWindow ()
{
  int  i, j, k;
  char *next;
  char ticker[7]; /* one more than expected size */
  char type[16];
  char server[16];
  char *item;

  for (i=0; i < saveCount; i++) {

    /* Get ticker symbol */
    XmStringGetLtoR ( save[i], XmFONTLIST_DEFAULT_TAG, &item);
    
    sscanf (item, "%6s =>%15c @%15c", ticker, type, server);
    type[15] = server[15] = '\0';     /* sscanf does not null terminate %c */
    if ((next = strrchr (type, ' '))) /* Prune trailing white space        */
      while (isspace(*next)) 
        *next-- = '\0';
    if ((next = strrchr (server, ' ')))
      while (isspace(*next)) 
        *next-- = '\0';

    /* Get server */
    /* 
    ** Only compares as much of server as we have preserved
    */
    for (j=0; j < numServer(); j++) {
      if (strncmp (server, getServerTitle(j), strlen(server)) == 0)
        break;
    }
    if (j == numServer()) {
      sprintf ( errmsg,
	        "Can not find server '%s', using first server.\n"
	        "This may cause server type values to not be found.",
		server);
      write_status (errmsg, WARN);
      j = 0;
    }
    /* Get server type */
    /* 
    ** Only compares as much of server type as we have preserved
    */
    for (k=0; k < numServerType(); k++) {
      setServerType(k);
      if (strncmp (type, getServer(TITLE), strlen(type)) == 0)
        break;
    }
    if (k == numServerType()) {
      sprintf ( errmsg,
	        "Can not find server type '%s', using first type.\n"
	        "This may cause table values to not be found.",
		type);
      write_status (errmsg, WARN);
      k = 0;
    }

    /* Make a structure with name and details for each ticker symbol */

    /* If ticker name changed reuse, else leave alone */
    if ( i < numTick && strcmp(ticker, tickTable[i]->name) ) {
      freePattern (tickTable[i]->detail);
      XtFree (tickTable[i]->name);

      tickTable[i]->name = XtNewString (ticker);
      tickTable[i]->type = k;
      tickTable[i]->server = j;
      tickTable[i]->detail = (QUERY_STRUCT *)NULL;
    }

    /* Add new ticker */
    if ( i >= numTick ) {
      tickTable = (TICK_STRUCT **) XtRealloc ( (char *)tickTable, 
                                               (i+1) * sizeof(TICK_STRUCT **) );
      tickTable[numTick++] = (TICK_STRUCT *) XtCalloc ( 1, sizeof(TICK_STRUCT));

      tickTable[i]->name = XtNewString (ticker);
      tickTable[i]->type = k;
      tickTable[i]->server = j;
      tickTable[i]->detail = (QUERY_STRUCT *)NULL;
    }
  }

  for (; i < numTick; i++) {                  /* Blow away extras */
    freePattern (tickTable[i]->detail);
    XtFree (tickTable[i]->name);
  }

  numTick = saveCount;
  detailUpdateMainWindow ();
  makeTextTape();
}


/* If symbol is in list, return its position, else return 0 */
static int tickIsDuplicate ( Widget list, char *symbol )
{
  XmStringTable ticklist;
  char *item;
  char name[7];
  int i, count;

  XtVaGetValues ( list, XmNitemCount, &count,
		        XmNitems,     &ticklist,
	          NULL);
  for ( i=0; i < count; i++) {
    XmStringGetLtoR ( ticklist[i], XmFONTLIST_DEFAULT_TAG, &item);
    sscanf ( item, "%6s", name );
    if (strcasecmp ( name, symbol) == 0 ) {
      XtFree(item);
      return (i+1);
    }
    XtFree(item);
  }
  return 0;  
}

/* Callback for ticker symbol dialog */
/* ARGSUSED */
static void procTick (Widget w, int client_data, XtPointer call_data)
{
  static Widget list = (Widget) NULL;
  static WidgetList field = (WidgetList)NULL;
  char *symbol[TICKFIELDS];

  XmStringTable orig;
  int origCount;

  if (list == (Widget) NULL) {
    list = XtNameToWidget (GetTopShell(w), 
                 "StockPane.TickForm.StockListSW.StockList");
    if (list) 
      /* So we can find the text fields to refill */
      XtVaGetValues (list, XmNuserData, &field, NULL);
  }

  switch (client_data) {
 
    case 0: /* Ok */
            XtPopdown (GetTopShell(w));

	    /* Save list contents */
	    if (saveCount) {
	      for (origCount=0; origCount < saveCount; origCount++)
	        XmStringFree (save[origCount]);
	      XtFree ((char *)save);
	      saveCount = 0;
	    }
	    XtVaGetValues (list, 
			   XmNitemCount, &saveCount, 
			   XmNitems, &orig, 
			   NULL);
	    if (saveCount) {
	      save = (XmStringTable) XtCalloc (saveCount, sizeof(XmString));
	      if (save) {
	        for (origCount=0; origCount < saveCount; origCount++)
	          save[origCount] = XmStringCopy (orig[origCount]);
	      } else
		saveCount = 0;
	    }

            /* Update main window table */
	    tickAddMainWindow ();

            break;

    case 1: /* Cancel */
            XtPopdown (GetTopShell(w));

	    /* Restore previous list */
	    XmListDeleteAllItems (list);
	    if (saveCount)
	      XmListAddItems (list, save, saveCount, 1);
            break;

    case 2: /* Add */
            {
              XmString listItem;
              char *line;

              /* Get ticker symbol */
              if (field[0]) {
                symbol[0] = XmTextFieldGetString (field[0]);

                /* If not blank or duplicate, add to list */
                if (strlen(symbol[0])) {
		  /* Not Blank */
	          if ( !tickIsDuplicate (list, symbol[0]) ) {
		    /* and Not Duplicate */
		    symbol[1] = getServer(TITLE);
		    symbol[2] = getServerTitle(curServer);

                    line = XtCalloc (80, sizeof (char));
                    sprintf (line, "%-6.6s =>%-15.15s @%-15.15s", 
			     symbol[0], symbol[1], symbol[2]);
                    listItem = XmStringCreateLocalized (line);
                    XmListAddItem (list, listItem, 0);
		    XmListSetBottomPos (list, 0);
                    XtFree (line);
                    XmStringFree (listItem);
                  } else {
		    /* and Duplicate */
		    sprintf (errmsg, "Symbol '%s' is already in the list.", 
			     symbol[0]);
		    write_status (errmsg, ERR);
		  }
	        } else {
		  /* Blank */
	          write_status ("Symbol is empty.", ERR);
	        }

                XtFree (symbol[0]);
                XmTextFieldSetSelection (field[0], 
                                         0, XmTextGetLastPosition(field[0]),
                                         CurrentTime);
              } else
	        write_status ("Internal error, no ticker text widget!", ERR);
            }
            break;

    case 3: /* Remove */
            {
              char line[80];
	      int pos;

              /* Get ticker symbol */
              if (field[0]) {
                symbol[0] = XmTextFieldGetString (field[0]);

                /* If not blank and exists, remove from list */
                if (strlen(symbol[0]) && 
	  	     (pos = tickIsDuplicate (list, symbol[0])) ) {
	          XmListDeletePos (list, pos); 
	        } else {
		  sprintf (line, "Symbol '%s' is not in the list.", symbol[0]);
		  write_status (line, ERR);
	        }

                XtFree (symbol[0]);
                XmTextFieldSetSelection (field[0], 
                                         0, XmTextGetLastPosition(field[0]),
                                         CurrentTime);
              } else
	        write_status ("Internal error, no ticker text widget!", ERR);
            }
            break;

    default: break;
  }

}


/* ARGSUSED */
static void textTick (Widget w, XtPointer client_data, XtPointer call_data)
{
  /* Simulate add button */
  procTick (w, 2, (XtPointer) NULL);
}


/* Read list params back into text fields when list item clicked on */
/* ARGSUSED */
static void loadTick (Widget w, XtPointer client_data, 
                      XmListCallbackStruct *call_data)
{
  int i;
  char *next;
  char field[TICKFIELDS][10];
  WidgetList textfield;

  /* Get the text field widgets */
  XtVaGetValues (w, XmNuserData, &textfield, NULL);

  XmStringGetLtoR ( call_data->item, XmFONTLIST_DEFAULT_TAG, &next );
  sscanf ( next, "%10c%10c%10c%10c%10c",         /* 9 for symbol, 1 space */
	   &(field[0][0]), &(field[1][0]), &(field[2][0]), 
           &(field[3][0]), &(field[4][0]) );

  /* Assume that the ticker symbol is never NULL or blanks */
  for ( i=0; i < TICKFIELDS; i++) {
    char temp[9];

    field[i][9] = '\0';                          /* end string */

    if (textfield[i]) {
      if (sscanf ( field[i], "%s", temp) == 1)   /* skip blanks */
        XmTextFieldSetString ( textfield[i], temp );
      else
        XmTextFieldSetString ( textfield[i], "" );
    }
  }
  XtFree (next);
}

/* ARGSUSED */
static void tickTypeCB ( Widget w, XtPointer client_data, XtPointer call_data)
{
  if ((int)client_data != curType) {
    curType = (int)client_data;
    setServerType (curType);
  }
}


/* ARGSUSED */
static void tickServerCB ( Widget w, XtPointer client_data, XtPointer call_data)
{
  if ((int)client_data != curServer) {
    Widget othermenu;
    XtVaGetValues (XtParent(w), XmNuserData, &othermenu, NULL);
    curServer = (int)client_data;
    setCurServer (curType);
    tickMakeTypeMenu(othermenu, 0);
  }
}

void tickMakeTypeMenu(Widget menu, int type)
{
  WidgetList buttons;
  int num;
  Widget button, othermenu;
  int i;

  if (menu == (Widget) NULL)
    return;

  /* Out with the old... */
  XtVaGetValues ( menu,
		  XmNchildren, &buttons,
		  XmNnumChildren, &num,
		  XmNuserData, &othermenu,
		  NULL );
  for (i=0; i < num; i++)
    XtDestroyWidget ( buttons[i] );

  /* ...in with the new. */
  if (type == 0 ) {       /* Server types */
    for (i = 0; i < numServerType(); i++) {
      setServerType(i);
      button = XtVaCreateManagedWidget (getServer(TITLE),
		                        xmPushButtonGadgetClass, menu, NULL);
      XtAddCallback (button, XmNactivateCallback, tickTypeCB, (XtPointer)i);
    }
    setServerType(0);
  } else if (type == 1) { /* Quote server */
    for (i = 0; i < numServer(); i++) {
      button = XtVaCreateManagedWidget (getServerTitle(i),
		     xmPushButtonGadgetClass, menu, NULL);
      XtAddCallback (button, XmNactivateCallback, tickServerCB, (XtPointer)i);
    }
    /* When we change servers, we need to refresh type list */
    if (getCurServer() != 0) {
      setCurServer(0);
      tickMakeTypeMenu(othermenu, 0);
    }
  }
}


Widget createTickDialog ()
{
  Widget dialog, form, pane, row, menu[2], list, button, option;
  static Widget text[] = { (Widget)NULL, (Widget)NULL, (Widget)NULL, 
                           (Widget)NULL, (Widget)NULL };

  Dimension width, height, border;

  int num;
  Arg args[10];

  char *labels[] = { "label_0", "label_1", "label_2"};
  char *texts[] = { "text_0"};
  char *buttons[] = { "button_0","button_1","button_2","button_3"};
  char *menus[] = { "TickServerMenu", "TickTypeMenu"};
  char *menupulldown[] = { "TickServerMenu_pulldown", "TickTypeMenu_pulldown"};


  dialog = XtVaCreatePopupShell ("OptionTick", 
                            xmDialogShellWidgetClass,
                            GetTopShell (Toplevel),
                            NULL);

  pane = XtVaCreateWidget ("StockPane", xmPanedWindowWidgetClass, 
                            dialog,
                            XmNsashWidth, 1,
                            XmNsashHeight, 1,
                            NULL);

  /* form to hold rowcolumn form, scrolled list */
  form = XtVaCreateWidget ("TickForm", xmFormWidgetClass, pane, NULL);

  row = XtVaCreateWidget ("TickRow", xmRowColumnWidgetClass, form,
                            XmNnavigationType, XmNONE,
                            XmNtopAttachment, XmATTACH_FORM,
                            XmNleftAttachment, XmATTACH_FORM,
                            XmNnumColumns, XtNumber(labels),
                            XmNorientation, XmHORIZONTAL,
                            XmNpacking, XmPACK_COLUMN,
                            NULL);
                
  for (num=0; num < XtNumber(labels); num++) {
    /* Label describing entry */
    XtVaCreateManagedWidget (labels[num], xmLabelGadgetClass, row, NULL);

    /* Text widget for number entry */
    if (num == 1) {
      text[0] = XtVaCreateManagedWidget (texts[0],
                            xmTextFieldWidgetClass, row,
                            XmNnavigationType, XmNONE,
                            NULL);
      XtAddCallback (text[0], XmNactivateCallback,
                     (XtCallbackProc) textTick, (XtPointer) 0);
    } else {
      if (num == 0) { /* Server */
        menu[0] = XmCreatePulldownMenu (row, menupulldown[0], args, 0);
        XtSetArg (args[0], XmNsubMenuId, menu[0]);
        option = XmCreateOptionMenu (row, menus[0], args, 1);
        tickMakeTypeMenu (menu[0], 1);
      } else {        /* Server types */
        menu[1] = XmCreatePulldownMenu (row, menupulldown[1], args, 0);
        XtSetArg (args[0], XmNsubMenuId, menu[1]);
        option = XmCreateOptionMenu (row, menus[1], args, 1);
        tickMakeTypeMenu (menu[1], 0);
        /* Save type menu in server menu, so it can be recreated */
        /* on server change                                      */
        XtVaSetValues (menu[0], XmNuserData, menu[1], NULL); 
      }
      XtManageChild (option);
    }
  }
  XtManageChild (row);

  /* List holding what's been added to date */
  num = 0;
  XtSetArg(args[num], XmNscrollBarDisplayPolicy, XmSTATIC); num++;
  XtSetArg(args[num], XmNvisibleItemCount, 5); num++;
  XtSetArg(args[num], XmNselectionPolicy,  XmSINGLE_SELECT); num++;
  XtSetArg(args[num], XmNtopAttachment,    XmATTACH_WIDGET); num++;
  XtSetArg(args[num], XmNtopWidget,        row); num++;
  XtSetArg(args[num], XmNleftAttachment,   XmATTACH_FORM); num++;
  XtSetArg(args[num], XmNrightAttachment,  XmATTACH_FORM); num++;
  /* Needed when we restore ticker symbols and then try to delete them 
     (loadTick). procTick is not called yet, so fields[] is not valid. */
  XtSetArg(args[num], XmNuserData, text); num++;
  list = XmCreateScrolledList (form, "StockList", args, num);

  XtAddCallback (list, XmNdefaultActionCallback,
                        (XtCallbackProc) loadTick, NULL);
  XtManageChild (list);
  XtManageChild (form);

  /* if we have restored items, load em up */
  if (saveCount)
    XmListAddItems (list, save, saveCount, 1);

  /* Buttons to add delete cancel exit */
  form = XtVaCreateWidget ("ButForm", xmFormWidgetClass, pane,
                            XmNfractionBase, XtNumber(buttons),
                            NULL);

  for (num=0; num < XtNumber(buttons); num++) {
    button = XtVaCreateManagedWidget (buttons[num], 
                            xmPushButtonWidgetClass, form,
                            XmNtopAttachment, XmATTACH_FORM,
                            XmNbottomAttachment, XmATTACH_FORM,
                            XmNleftAttachment, XmATTACH_POSITION,
                            XmNleftPosition, num,
                            XmNrightAttachment, XmATTACH_POSITION,
                            XmNrightPosition, num+1,
                            XmNshowAsDefault, (num==0)?True:False,
                            XmNdefaultButtonShadowThickness, 1,
                            NULL);
    XtAddCallback (button, XmNactivateCallback,
                            (XtCallbackProc) procTick, (XtPointer)num);
  }
  XtManageChild (form);
  XtManageChild (pane);

  /* Prevent pane from changing size */
  XtVaGetValues (dialog, 
                 XmNwidth, &width,
                 XmNheight, &height,
                 XmNborderWidth, &border,
                 NULL);

  XtVaSetValues (dialog,
                 XmNminWidth,  width +  border,
                 XmNmaxWidth,  width +  border,
                 XmNminHeight, height + border,
                 XmNmaxHeight, height + border,
                 NULL);
  return (dialog);
}
